from typing import Callable, Dict, Sequence

import torch as t
from torch.utils.data import DataLoader

MetricFunc = Callable[[t.Tensor, t.Tensor], float]
DEVICE = t.device("cuda" if t.cuda.is_available() else "cpu")


def evaluate(
    model: t.nn.Module, dataloader: DataLoader, metric_functions: Sequence[MetricFunc]
) -> Dict[str, float]:
    model.eval()

    _, tmp = next(iter(dataloader))
    target_dtype = tmp.dtype

    outputs = t.Tensor([])  # pyre-ignore
    targets = t.Tensor([]).to(target_dtype)  # pyre-ignore

    with t.no_grad():
        for batch_inputs, batch_targets in dataloader:
            batch_inputs = batch_inputs.to(DEVICE)
            batch_targets = batch_targets.to(DEVICE)

            batch_outputs = model(batch_inputs)
            outputs = t.cat((outputs, batch_outputs))
            targets = t.cat((targets, batch_targets))

    metrics: Dict[str, float] = {}
    for metric_func in metric_functions:
        metrics[metric_func.__name__] = metric_func(targets, outputs)

    return metrics



import logging
import sys
from typing import Any, Callable, Dict, Optional, Tuple, cast

import torch as t
from ax.service.managed_loop import optimize
from torch.utils.data import Dataset

from .hyperparams import Hyperparams, HyperparamsSpec
from .ml_logger import MLExperiment
from .trainer import Trainer, TrainerArgs

ArgsBuilderType = Callable[[Hyperparams, Dataset, Dataset], TrainerArgs]
MetricFunc = Callable[[t.Tensor, t.Tensor], float]
HparamsFactory = Callable[[Dict[str, Any]], Hyperparams]

logger: logging.Logger = logging.getLogger(__name__)


class Tuner:
    def __init__(
        self,
        experiment: MLExperiment,
        trainset: Dataset,
        valset: Dataset,
        obj_metric: MetricFunc,
        minimize: bool = False,
    ) -> None:
        self._obj_metric: MetricFunc = obj_metric
        self._trainer = Trainer(experiment, trainset, valset, [obj_metric])
        self._trainer.metrics_log_frequency = sys.maxsize
        self._model_log_frequency: int = sys.maxsize
        self._minimize = minimize

        self._hparams_factory: Optional[HparamsFactory] = None
        self._args_builder: Optional[ArgsBuilderType] = None

    @property
    def metrics_log_frequency(self) -> int:
        return self._trainer.metrics_log_frequency

    @metrics_log_frequency.setter
    def metrics_log_frequency(self, freq: int) -> None:
        self._trainer.metrics_log_frequency = freq

    def _train_eval(self, hyper_params: Dict[str, Any]) -> Dict[str, Tuple[float, float]]:
        hparams_factory = cast(HparamsFactory, self._hparams_factory)
        hparams: Hyperparams = hparams_factory(**hyper_params)

        args_builder = cast(ArgsBuilderType, self._args_builder)
        self._trainer.train(hparams, args_builder)
        obj_name = self._obj_metric.__name__
        obj_val: Tuple[float, float] = (self._trainer.final_metrics["val_" + obj_name], 0.0)
        return {obj_name: obj_val}

    def tune(
        self, hparams_spec: HyperparamsSpec, args_builder: ArgsBuilderType, total_trials: int = 20
    ) -> Hyperparams:
        self._hparams_factory = hparams_spec.factory
        self._args_builder = args_builder

        best_params, values, _, _ = optimize(
            hparams_spec.spec,
            evaluation_function=self._train_eval,  # pyre-ignore
            objective_name=self._obj_metric.__name__,
            total_trials=total_trials,
            minimize=self._minimize,
        )

        logger.info(f"best_params={best_params} values={values}")
        return hparams_spec.factory(**best_params)  # pyre-ignore



import pprint
import uuid
from dataclasses import dataclass
from typing import Callable, Sequence, Tuple

import sklearn.datasets as skdata
import torch as t
from sklearn.metrics import mean_absolute_error
from torch.utils.data import DataLoader, Dataset, TensorDataset

from ..evaluator import evaluate
from ..hyperparams import Hyperparams
from ..ml_loggers.stdout_logger import StdoutMLExperiment
from ..trainer import Trainer, TrainerArgs

MetricFunc = Callable[[t.Tensor, t.Tensor], float]


class Regressor(t.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.fc1 = t.nn.Linear(5, 3)
        self.fc2 = t.nn.Linear(3, 1)

    def forward(self, batch_x: t.Tensor) -> t.Tensor:
        x = t.nn.functional.relu(self.fc1(batch_x))
        batch_y_hat = self.fc2(x)
        return t.squeeze(batch_y_hat, dim=1)


@dataclass
class MyHyperparams(Hyperparams):
    batch_size: int
    n_epochs: int
    lr: float


def gen_datasets(n_samples: int) -> Tuple[Dataset, Dataset, Dataset]:
    all_X, all_y = skdata.make_regression(n_samples=n_samples, n_features=5, noise=0.5)

    train_size = int(n_samples * 0.7)
    val_size = int(n_samples * 0.2)

    train_X = all_X[:train_size]
    train_y = all_y[:train_size]
    trainset = TensorDataset(
        t.from_numpy(train_X).to(t.float32), t.from_numpy(train_y).to(t.float32)
    )

    val_X = all_X[train_size : train_size + val_size]
    val_y = all_y[train_size : train_size + val_size]
    valset = TensorDataset(t.from_numpy(val_X).to(t.float32), t.from_numpy(val_y).to(t.float32))

    test_X = all_X[train_size + val_size :]
    test_y = all_y[train_size + val_size :]
    testset = TensorDataset(t.from_numpy(test_X).to(t.float32), t.from_numpy(test_y).to(t.float32))

    return trainset, valset, testset


def build_trainer(hparams: MyHyperparams, trainset: Dataset, valset: Dataset) -> TrainerArgs:
    run_name = "run-" + str(uuid.uuid4())[:8]
    model = Regressor()
    optim = t.optim.Adam(model.parameters(), lr=hparams.lr)
    loss_fn = t.nn.MSELoss(reduction="mean")
    traindl = DataLoader(trainset, batch_size=hparams.batch_size, shuffle=True)
    valdl = DataLoader(valset, batch_size=5000)
    return TrainerArgs(
        run_name=run_name,
        model=model,
        optim=optim,
        loss_fn=loss_fn,
        trainloader=traindl,
        valloader=valdl,
        n_epochs=hparams.n_epochs,
    )


def train(
    hparams: MyHyperparams,
    trainset: Dataset,
    valset: Dataset,
    metric_functions: Sequence[MetricFunc],
) -> None:
    exp = StdoutMLExperiment("regress-exp")
    trainer = Trainer(exp, trainset, valset, metric_functions=metric_functions)
    trainer.metrics_log_frequency = 1
    trainer.model_log_frequency = 5
    trainer.train(hparams, build_trainer)  # pyre-ignore


def main() -> None:
    hparams = MyHyperparams(batch_size=16, n_epochs=10, lr=0.03)
    trainset, valset, testset = gen_datasets(n_samples=100_000)
    train(hparams, trainset, valset, [mean_absolute_error])

    print("Enter the path of the saved model:")
    model_path = input()
    model = t.load(model_path)
    testdl = DataLoader(testset, batch_size=5000)
    test_metrics = evaluate(model, testdl, [mean_absolute_error])
    pprint.pprint(test_metrics)


if __name__ == "__main__":
    main()



import logging
import sys
from collections import namedtuple
from dataclasses import dataclass
from typing import Callable, Dict, Optional, Sequence, Tuple, cast

import numpy as np
import torch as t
from torch.optim import Optimizer
from torch.utils.data import DataLoader, Dataset

from .hyperparams import Hyperparams
from .ml_logger import MLExperiment

MetricFunc = Callable[[t.Tensor, t.Tensor], float]
logger = logging.getLogger(__name__)

Metric = namedtuple("Metric", ["name", "value"])


@dataclass
class TrainerArgs:
    run_name: str
    model: t.nn.Module
    optim: Optimizer
    loss_fn: Callable[[t.Tensor, t.Tensor], t.Tensor]
    trainloader: DataLoader
    valloader: DataLoader
    n_epochs: int
    grad_warning_threshold: float = 1000
    loss_warning_threshold: float = 10_000
    clip_grads: float = float("inf")


ArgsBuilderType = Callable[[Hyperparams, Dataset, Dataset], TrainerArgs]

DEVICE = t.device("cuda" if t.cuda.is_available() else "cpu")


class Trainer:
    def __init__(
        self,
        experiment: MLExperiment,
        trainset: Dataset,
        valset: Dataset,
        metric_functions: Sequence[MetricFunc],
    ) -> None:
        self._trainset = trainset
        self._valset = valset

        tempdl = DataLoader(self._trainset, batch_size=2)
        _, batch_target = next(iter(tempdl))
        self._target_dtype = batch_target.dtype

        self._metric_functions = metric_functions
        self._experiment = experiment
        self._args: Optional[TrainerArgs] = None

        self.metrics_log_frequency: int = 1
        self.model_log_frequency: int = sys.maxsize

        self.model: Optional[t.nn.Module] = None
        self.final_metrics: Dict[str, float] = {}

    def _train(self, calc_metrics: bool = False,) -> Tuple[float, Dict[str, float]]:
        args = cast(TrainerArgs, self._args)
        model = cast(t.nn.Module, self.model)

        traindl = args.trainloader
        optim = args.optim
        loss_fn = args.loss_fn
        clip_grads = args.clip_grads

        model.train()
        losses = []
        outputs = t.Tensor([])  # pyre-ignore
        targets = t.Tensor([]).to(self._target_dtype)  # pyre-ignore
        with t.enable_grad():
            for batch_inputs, batch_targets in traindl:
                batch_inputs = batch_inputs.to(DEVICE)
                batch_targets = batch_targets.to(DEVICE)

                optim.zero_grad()
                batch_outputs = model.forward(batch_inputs)
                loss = loss_fn(batch_outputs, batch_targets)
                loss.backward()

                if clip_grads < float("inf"):
                    t.nn.utils.clip_grad_value_(model.parameters(), clip_grads)

                optim.step()

                losses.append(loss.detach().item())

                if calc_metrics:
                    outputs = t.cat((outputs, batch_outputs.detach()))
                    targets = t.cat((targets, batch_targets))

        loss = np.mean(losses)

        metrics: Dict[str, float] = {}
        if calc_metrics:
            metrics["train_loss"] = loss
            for metric_func in self._metric_functions:
                metrics["train_" + metric_func.__name__] = metric_func(targets, outputs)

        return loss, metrics

    def _validate(self):
        args = cast(TrainerArgs, self._args)
        model = cast(t.nn.Module, self.model)

        valdl = args.valloader
        loss_fn = args.loss_fn

        model.eval()
        losses = []
        outputs = t.Tensor([])
        targets = t.Tensor([]).to(self._target_dtype)
        with t.no_grad():
            for batch_inputs, batch_targets in valdl:
                batch_inputs = batch_inputs.to(DEVICE)
                batch_targets = batch_targets.to(DEVICE)
                batch_outputs = model(batch_inputs)
                loss = loss_fn(batch_outputs, batch_targets)
                losses.append(loss.detach().item())
                outputs = t.cat((outputs, batch_outputs))
                targets = t.cat((targets, batch_targets))

        loss = np.mean(losses)

        metrics: Dict[str, float] = {}
        metrics["val_loss"] = loss
        for metric_func in self._metric_functions:
            metrics["val_" + metric_func.__name__] = metric_func(targets, outputs)

        return loss, metrics

    def train(self, hparams: Hyperparams, args_builder: ArgsBuilderType) -> None:
        args = args_builder(hparams, self._trainset, self._valset)
        model = args.model.to(DEVICE)

        self._args = args
        self.model = model

        with self._experiment.start_run(args.run_name) as run:
            for epoch in range(1, args.n_epochs + 1):
                train_loss: float = 0.0
                val_loss: float = 0.0
                if epoch % self.metrics_log_frequency == 0:
                    train_loss, metrics = self._train(calc_metrics=True)
                    val_loss, val_metrics = self._validate()
                    metrics.update(val_metrics)
                    for metric_name, metric_value in metrics.items():
                        run.log_metric(epoch, metric_name, metric_value)
                else:
                    train_loss, _ = self._train()

                # Warn about high loss values
                if train_loss > args.loss_warning_threshold:
                    logger.warning(f"Epoch {epoch}: Training loss {train_loss:.3f} is too high!")
                if val_loss > args.loss_warning_threshold:
                    logger.warning(f"Epoch {epoch}: Validation loss {val_loss:.3f} is too high!")

                # Warn about high gradients
                # pyre-ignore
                max_grad = t.max(t.Tensor([t.max(param) for param in self.model.parameters()]))
                if max_grad > args.grad_warning_threshold:
                    logger.warning(f"Epoch {epoch}: Gradient {max_grad:.3f} is too high!")

                # Checkpoint the model if needed
                if epoch % self.model_log_frequency == 0:
                    run.log_model(epoch, model)

            # Calculate the final metrics on the validation dataset and log them
            # as part of run summary along with the hyperparams
            _, self.final_metrics = self._validate()
            run.log_summary(hparams.to_dict(), self.final_metrics)



import dataclasses as dc
import os.path as path
from configparser import ConfigParser
from typing import Any, Callable, Dict, List


@dc.dataclass
class Hyperparams:
    def to_dict(self):
        ret = {}
        for fld in dc.fields(self):
            ret[fld.name] = getattr(self, fld.name)
        return ret

    @classmethod
    def load(cls, conffile: str) -> "Hyperparams":
        conf = ConfigParser()
        if not path.exists(conffile):
            raise ValueError(f"Unable to find {conffile}")
        conf.read(conffile)
        confgetters = {int: conf.getint, float: conf.getfloat, bool: conf.getboolean, str: conf.get}
        kwargs = {}
        for fld in dc.fields(cls):
            if fld.name in conf["HYPERPARAMS"]:
                kwargs[fld.name] = confgetters[fld.type]("HYPERPARAMS", fld.name)
        return cls(**kwargs)


@dc.dataclass
class HyperparamsSpec:
    spec: List[Dict[str, Any]]
    factory: Callable[[Dict[str, Any]], Hyperparams]



from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, Dict, Generator

import torch


class MLRun(ABC):
    def __init__(self, name: str) -> None:
        self.name = name

    @abstractmethod
    def log_summary(self, hparams: Dict[str, Any], metrics: Dict[str, float]) -> None:
        raise NotImplementedError("Subclass must implement!")

    @abstractmethod
    def log_metric(self, step: int, name: str, val: float) -> None:
        raise NotImplementedError("Subclass must implement!")

    @abstractmethod
    def log_params(self, step: int, state: Dict[str, torch.Tensor]) -> None:
        raise NotImplementedError("Subclass must implement!")

    @abstractmethod
    def log_model(self, step, model) -> None:
        raise NotImplementedError("Subclass must implement!")


class MLExperiment(ABC):
    def __init__(self, name: str) -> None:
        self.name = name

    @abstractmethod
    @contextmanager
    def start_run(self, name: str) -> Generator[MLRun, None, None]:
        pass



import logging
import pprint
import uuid
from dataclasses import dataclass
from typing import Callable, Sequence, Tuple

import sklearn.datasets as skdata
import torch as t
from sklearn.metrics import accuracy_score
from torch.utils.data import DataLoader, Dataset, TensorDataset

from ..evaluator import evaluate
from ..hyperparams import Hyperparams
from ..ml_loggers.mlflow_logger import MLFlowMLExperiment
from ..trainer import Trainer, TrainerArgs

logging.basicConfig(level=logging.DEBUG)
MetricFunc = Callable[[t.Tensor, t.Tensor], float]


class BinaryClassifier(t.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.fc1 = t.nn.Linear(20, 8)
        self.fc2 = t.nn.Linear(8, 1)

    def forward(self, batch_x: t.Tensor) -> t.Tensor:
        x = t.nn.functional.relu(self.fc1(batch_x))
        batch_y_hat = t.sigmoid(self.fc2(x))
        return t.squeeze(batch_y_hat, dim=1)


@dataclass
class MyHyperparams(Hyperparams):
    n_epochs: int
    batch_size: int
    lr: float
    true_cutoff: float


def gen_datasets(n_samples: int) -> Tuple[Dataset, Dataset, Dataset]:
    X, y = skdata.make_classification(
        n_samples=n_samples,
        n_features=20,
        n_informative=10,
        n_redundant=7,
        n_repeated=3,
        n_classes=2,
        flip_y=0.05,  # larger values make the task hard
        class_sep=0.5,  # larger values makes the task easy
        random_state=10,
    )

    train_size = int(n_samples * 0.7)
    val_size = int(n_samples * 0.2)

    train_X = X[:train_size]
    train_y = y[:train_size]
    trainset = TensorDataset(
        t.from_numpy(train_X).to(t.float32), t.from_numpy(train_y).to(t.float32)
    )

    val_X = X[train_size : train_size + val_size]
    val_y = y[train_size : train_size + val_size]
    valset = TensorDataset(t.from_numpy(val_X).to(t.float32), t.from_numpy(val_y).to(t.float32))

    test_X = X[train_size + val_size :]
    test_y = y[train_size + val_size :]
    testset = TensorDataset(t.from_numpy(test_X).to(t.float32), t.from_numpy(test_y).to(t.float32))

    return trainset, valset, testset


def build_accuracy(cutoff: float) -> Callable[[t.Tensor, t.Tensor], float]:
    def accuracy(y_true, y_hat):
        y_pred = y_hat > cutoff
        return accuracy_score(y_true, y_pred)

    return accuracy


def build_trainer(hparams: MyHyperparams, trainset: Dataset, valset: Dataset) -> TrainerArgs:
    run_name = "run-" + str(uuid.uuid4())[:8]
    print(f"Starting run {run_name}")
    model = BinaryClassifier()
    optim = t.optim.Adam(model.parameters(), lr=hparams.lr)
    loss_fn = t.nn.BCELoss()
    traindl = DataLoader(trainset, batch_size=hparams.batch_size, shuffle=True)
    valdl = DataLoader(valset, batch_size=5000)
    return TrainerArgs(
        run_name=run_name,
        model=model,
        optim=optim,
        loss_fn=loss_fn,
        trainloader=traindl,
        valloader=valdl,
        n_epochs=hparams.n_epochs,
    )


def train(
    hparams: MyHyperparams,
    trainset: Dataset,
    valset: Dataset,
    metric_functions: Sequence[MetricFunc],
) -> None:
    # Do the following to start mlflow properly.
    # cd ~/temp/mlflow
    # mlflow ui
    # This will create a directory called mlruns under ~/temp/mlflow/mlruns and start the mlflow webserver on port 5000.
    # Now set the experiment dir to ~temp/mlflow/mlruns
    exp = MLFlowMLExperiment("binaryclass-exp", "~/temp/mlflow/mlruns")
    trainer = Trainer(exp, trainset, valset, metric_functions=metric_functions)
    trainer.metrics_log_frequency = 1
    trainer.model_log_frequency = 5
    trainer.train(hparams, build_trainer)  # pyre-ignore


def main() -> None:
    hparams = MyHyperparams(n_epochs=7, lr=0.005, batch_size=32, true_cutoff=0.5)
    trainset, valset, testset = gen_datasets(n_samples=100_000)
    accuracy = build_accuracy(cutoff=hparams.true_cutoff)
    train(hparams, trainset, valset, [accuracy])

    print("Enter the path of the saved model:")
    model_path = input()
    model = t.load(model_path)
    testdl = DataLoader(testset, batch_size=5000)
    test_metrics = evaluate(model, testdl, [accuracy])
    pprint.pprint(test_metrics)


if __name__ == "__main__":
    main()



import logging
import os
from contextlib import contextmanager
from typing import Any, Dict, Generator, Optional

import mlflow
import mlflow.pytorch
import torch

from ..ml_logger import MLExperiment, MLRun

logger = logging.getLogger(__name__)


class MLFlowMLRun(MLRun):
    def __init__(self, name: str, mlflow_run: mlflow.ActiveRun) -> None:
        super().__init__(name)
        self._run: mlflow.ActiveRun = mlflow_run
        self._last_step = 0

    def log_summary(self, hparams: Dict[str, Any], metrics: Dict[str, float]) -> None:
        mlflow.log_params(hparams)
        for metric_name, metric_val in metrics.items():
            self.log_metric(self._last_step + 1, metric_name, metric_val)

    def log_metric(self, step: int, name: str, val: float) -> None:
        mlflow.log_metric(name, val, step)
        self._last_step = step

    def log_params(self, step: int, state: Dict[str, torch.Tensor]) -> None:
        logger.warning("MLFlow does not support logging parameters.")

    def log_model(self, step, model) -> None:
        mlflow.pytorch.log_model(model, "model")
        self._last_step = step


class MLFlowMLExperiment(MLExperiment):
    def __init__(self, name: str, dirpath: Optional[str] = None) -> None:
        super().__init__(name)
        if dirpath:
            mlflow.set_tracking_uri(os.path.expanduser(dirpath))
        mlflow.set_experiment(name)

    @contextmanager
    def start_run(self, name: str) -> Generator[MLRun, None, None]:
        with mlflow.start_run(run_name=name) as mlflow_run:
            run = MLFlowMLRun(name, mlflow_run)
            yield run



import os.path as path
import pprint
import sys
import uuid
from dataclasses import dataclass
from typing import Callable, Tuple

import sklearn.datasets as skdata
import torch as t
from sklearn.metrics import accuracy_score
from torch.utils.data import DataLoader, Dataset, TensorDataset

from ..evaluator import evaluate
from ..hyperparams import Hyperparams, HyperparamsSpec
from ..ml_loggers.csv_logger import CsvMLExperiment
from ..ml_loggers.stdout_logger import StdoutMLExperiment
from ..trainer import Trainer, TrainerArgs
from ..tuner import Tuner

MetricFunc = Callable[[t.Tensor, t.Tensor], float]


class MulticlassClassifier(t.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.fc1 = t.nn.Linear(20, 16)
        self.fc2 = t.nn.Linear(16, 8)
        self.fc3 = t.nn.Linear(8, 5)

    def forward(self, batch_x: t.Tensor) -> t.Tensor:
        x = t.nn.functional.relu(self.fc1(batch_x))
        x = t.nn.functional.relu(self.fc2(x))
        batch_y_hat = t.nn.functional.relu(self.fc3(x))
        return batch_y_hat


@dataclass
class MyHyperparams(Hyperparams):
    batch_size: int
    n_epochs: int
    lr: float


def gen_datasets(n_samples: int) -> Tuple[Dataset, Dataset, Dataset]:
    X, y = skdata.make_classification(
        n_samples=n_samples,
        n_features=20,
        n_informative=10,
        n_redundant=7,
        n_repeated=3,
        n_classes=5,
        flip_y=0.05,  # larger values make the task hard
        class_sep=0.5,  # larger values makes the task easy
        random_state=10,
    )

    train_size = int(n_samples * 0.7)
    val_size = int(n_samples * 0.2)

    train_X = X[:train_size]
    train_y = y[:train_size]
    trainset = TensorDataset(t.from_numpy(train_X).to(t.float32), t.from_numpy(train_y).to(t.int64))

    val_X = X[train_size : train_size + val_size]
    val_y = y[train_size : train_size + val_size]
    valset = TensorDataset(t.from_numpy(val_X).to(t.float32), t.from_numpy(val_y).to(t.int64))

    test_X = X[train_size + val_size :]
    test_y = y[train_size + val_size :]
    testset = TensorDataset(t.from_numpy(test_X).to(t.float32), t.from_numpy(test_y).to(t.int64))

    return trainset, valset, testset


def accuracy(y_true: t.Tensor, y_hat: t.Tensor) -> float:
    y_pred = t.argmax(y_hat, dim=1)
    return accuracy_score(y_true, y_pred)


def build_trainer(hparams: MyHyperparams, trainset: Dataset, valset: Dataset,) -> TrainerArgs:
    run_name = "run-" + str(uuid.uuid4())[:8]
    print(f"\nStarting run {run_name}")
    model = MulticlassClassifier()
    optim = t.optim.Adam(model.parameters(), lr=hparams.lr)
    loss_fn = t.nn.CrossEntropyLoss()
    traindl = DataLoader(trainset, batch_size=hparams.batch_size, shuffle=True)
    valdl = DataLoader(valset, batch_size=1000)
    return TrainerArgs(
        run_name=run_name,
        model=model,
        optim=optim,
        loss_fn=loss_fn,
        trainloader=traindl,
        valloader=valdl,
        n_epochs=hparams.n_epochs,
    )


def train() -> None:
    hparams = MyHyperparams(batch_size=64, n_epochs=10, lr=0.01)
    trainset, valset, testset = gen_datasets(100_000)
    exp = CsvMLExperiment("multiclass-exp", "~/temp/experiments")
    trainer = Trainer(exp, trainset, valset, metric_functions=[accuracy])
    trainer.metrics_log_frequency = 1
    trainer.model_log_frequency = 5
    trainer.train(hparams, build_trainer)  # pyre-ignore

    print("Enter the path of the saved model:")
    model_path = input()
    if model_path and path.exists(model_path):
        model = t.load(model_path)
    else:
        print("Using trained model")
        model = trainer.model
    testdl = DataLoader(testset, batch_size=5000)
    test_metrics = evaluate(model, testdl, [accuracy])
    pprint.pprint(test_metrics)


def tune() -> None:
    exp = CsvMLExperiment("multiclass-exp-tune", "~/temp/experiments", stdout=False)
    trainset, valset, testset = gen_datasets(100_000)
    tuner = Tuner(exp, trainset, valset, accuracy)
    tuner.metrics_log_frequency = 1

    hparams_spec = HyperparamsSpec(
        factory=MyHyperparams,  # pyre-ignore
        spec=[
            {"name": "batch_size", "type": "choice", "value_type": "int", "values": [16, 32, 64]},
            {"name": "n_epochs", "type": "range", "value_type": "int", "bounds": [7, 23]},
            {"name": "lr", "type": "range", "bounds": [1e-4, 0.4], "log_scale": True},
        ],
    )
    # pyre-ignore
    best_params = tuner.tune(hparams_spec, build_trainer, total_trials=10)
    print(best_params)

    ans = input("Train using best params [Y/n]?")
    if ans.lower() != "n":
        exp = StdoutMLExperiment("best-model")
        trainer = Trainer(exp, trainset, valset, [accuracy])
        trainer.train(best_params, build_trainer)  # pyre-ignore
        best_model = trainer.model
        t.save(best_model, "best_model.pkl")
        print(trainer.final_metrics)


def main() -> None:
    if len(sys.argv) < 2 or sys.argv[1] not in ["train", "tune"]:
        print("Usage: python multiclass_classification_csv.py [train|tune]")
        sys.exit(1)

    if sys.argv[1] == "train":
        train()
    elif sys.argv[1] == "tune":
        tune()


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages


setup(
    name="avilabs-torchutils",
    version="3.0.0",
    description="Convenience utils for using pytorch",
    author="Avilay Parekh",
    author_email="avilay@gmail.com",
    license="MIT",
    url="https://gitlab.com/avilay/torchutils",
    packages=find_packages(),
    install_requires=["torch", "numpy", "sklearn", "ax-platform"],
)



import os
from contextlib import contextmanager
from pprint import pprint
from typing import Any, Dict, Generator, List

import torch

from ..ml_logger import MLExperiment, MLRun


class StdoutMLRun(MLRun):
    def __init__(self, name: str) -> None:
        super().__init__(name)
        self._models_root: str = os.path.join(".", "models")
        os.makedirs(self._models_root, exist_ok=True)
        self._curr_step = 0
        self._buffer: List[str] = []

    def log_summary(self, hparams: Dict[str, Any], metrics: Dict[str, float]) -> None:
        summary = {"hyperparams": hparams, "evaluation_metrics": {}}
        for metric_name, metric_value in metrics.items():
            summary["evaluation_metrics"][metric_name] = round(metric_value, 3)
        print("\nSummary:")
        pprint(summary)

    def log_metric(self, step: int, name: str, val: float) -> None:
        if step == self._curr_step:
            self._buffer.append(f"{name} = {val:.3f}")
        else:
            self._print_buffer()
            self._buffer = [f"{name} = {val:.3f}"]
            self._curr_step = step

    def log_params(self, step: int, state: Dict[str, torch.Tensor]) -> None:
        print("[WARN] Stdout does not support logging parameters")

    def log_model(self, step: int, model: torch.nn.Module) -> None:
        model_filename = f"{self.name}-model-{step}.pkl"
        model_path = os.path.join(self._models_root, model_filename)
        print(f"\nStep {step}: Checkpointing model at {model_filename}")
        torch.save(model, model_path)

    def close(self):
        self._print_buffer()

    def _print_buffer(self):
        if self._buffer:
            print(f"\nStep {self._curr_step}")
        for buf in self._buffer:
            print(f"\t{buf}")


class StdoutMLExperiment(MLExperiment):
    def __init__(self, name: str) -> None:
        super().__init__(name)

    @contextmanager
    def start_run(self, name: str) -> Generator[MLRun, None, None]:
        run = StdoutMLRun(name)
        print(f"Starting run {name}")
        yield run
        run.close()



import json
import logging
import os
import os.path as path
from contextlib import contextmanager
from datetime import datetime
from typing import Any, Dict, Generator

import numpy as np
import torch

from ..ml_logger import MLExperiment, MLRun
from .stdout_logger import StdoutMLRun

logger: logging.Logger = logging.getLogger(__name__)


class CsvMLRun(MLRun):
    def __init__(self, name: str, exproot: str) -> None:
        super().__init__(name)
        runroot = path.join(exproot, name.strip().lower())
        if path.exists(runroot):
            raise RuntimeError(f"{runroot} already exists! Choose a different run name.")
        os.mkdir(runroot)

        self._models_root: str = path.join(runroot, "models")
        os.mkdir(self._models_root)
        self._model_loc_file = open(path.join(runroot, "models_loc.csv"), "wt")
        print("timestamp,step,model_path", file=self._model_loc_file)

        self._summary_filename: str = path.join(runroot, "summary.json")

        self._metricsfile = open(path.join(runroot, "metrics.csv"), "wt")
        print("timestamp,step,name,value", file=self._metricsfile)

        # timestamp,bin_0_lower,bin_0_upper,bin_0_mid,bin_0_freq,bin_1_lower,bin_1_upper,...
        self._paramsfile = open(path.join(runroot, "params_hist.csv"), "wt")
        header = ["timestamp", "step", "param_name"]
        for i in range(10):
            header += [f"bin_{i}_lower", f"bin_{i}_upper", f"bin_{i}_mid", f"bin_{i}_count"]
        print(",".join(header), file=self._paramsfile)

    def log_summary(self, hparams: Dict[str, Any], metrics: Dict[str, float]) -> None:
        if path.exists(self._summary_filename):
            raise RuntimeError(f"{self._summary_filename} was already written once!")
        summary = {"hyperparams": hparams, "evaluation_metrics": metrics}
        with open(self._summary_filename, "wt") as f:
            json.dump(summary, f)

    def log_metric(self, step: int, name: str, val: float) -> None:
        print(f"{datetime.now()},{step},{name},{val:.6f}", file=self._metricsfile)

    def log_params(self, step: int, state: Dict[str, torch.Tensor]) -> None:
        ts = datetime.now()
        for param_name, params in state.items():
            params = torch.flatten(params)
            hist, bin_edges = np.histogram(params)
            record = f"{ts},{step},{param_name}"
            for i, freq in enumerate(hist):
                bin_mid = (bin_edges[i] + bin_edges[i + 1]) / 2
                record += f",{bin_edges[i]},{bin_edges[i+1]},{bin_mid},{freq}"
            print(record, file=self._paramsfile)

    def log_model(self, step: int, model: torch.nn.Module) -> None:
        model_filename = f"model-{step}.pkl"
        model_path = path.join(self._models_root, model_filename)
        print(f"{datetime.now()},{step},{model_filename}", file=self._model_loc_file)
        torch.save(model, model_path)

    def close(self) -> None:
        self._metricsfile.close()
        self._paramsfile.close()
        self._model_loc_file.close()


class CsvStdoutMLRun(MLRun):
    def __init__(self, name: str, exproot: str) -> None:
        super().__init__(name)
        self._csv_run = CsvMLRun(name, exproot)
        self._stdout_run = StdoutMLRun(name)

    def log_summary(self, hparams: Dict[str, Any], metrics: Dict[str, float]) -> None:
        self._csv_run.log_summary(hparams, metrics)
        self._stdout_run.log_summary(hparams, metrics)

    def log_metric(self, step: int, name: str, val: float) -> None:
        self._csv_run.log_metric(step, name, val)
        self._stdout_run.log_metric(step, name, val)

    def log_params(self, step: int, state: Dict[str, torch.Tensor]) -> None:
        self._csv_run.log_params(step, state)
        self._stdout_run.log_params(step, state)

    def log_model(self, step: int, model: torch.nn.Module) -> None:
        # No need to save the model twice. Just use one of the runs for this.
        print(f"\nStep {step}: Checkpointing model")
        self._csv_run.log_model(step, model)

    def close(self) -> None:
        self._csv_run.close()
        self._stdout_run.close()


class CsvMLExperiment(MLExperiment):
    def __init__(self, name: str, dirpath: str, stdout: bool = True) -> None:
        super().__init__(name)
        dirpath = path.expanduser(dirpath)
        self._exproot: str = path.join(dirpath, name.strip().lower())
        if path.exists(self._exproot):
            logger.warning(
                f"{self._exproot} already exists. Will add runs to the existing experiment."
            )
        else:
            os.mkdir(self._exproot)

        self._stdout = stdout

    @contextmanager
    def start_run(self, name: str) -> Generator[MLRun, None, None]:
        if self._stdout:
            run = CsvStdoutMLRun(name, self._exproot)
        else:
            run = CsvMLRun(name, self._exproot)
        yield run
        run.close()
